/**
 * file: tests/api/v0/productTag_test.go
 * author: Theo Technicguy
 * license: Apache 2.0
 *
 * Product tags APIv0 endpoint tests
 */

package v0_test

import (
	"bytes"
	"encoding/json"
	"fmt"
	"net/http"
	"testing"

	"git.licolas.net/hoffman/server/types"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

var (
	testProductTags = []*types.Tag{
		{Id: 1, Name: "tag1", Color: 0},
		{Id: 2, Name: "tag2", Color: 123456},
		{Id: 3, Name: "tag3", Color: 16581375},
	}
)

func testRequestTag[T any](
	t *testing.T,
	tss *testServerSetup,
	method string,
	url string,
	tag any,
	expect T,
	statusCode int,
) {
	testRequestWithMarshal[T, *types.Tag](t, tss, method, url, tag, expect, statusCode)
}

func testGetTag[T any](
	t *testing.T,
	tss *testServerSetup,
	url string,
	expect T,
	statusCode int,
) {
	method := "GET"
	testRequestTag[T](t, tss, method, url, nil, expect, statusCode)
}

func testPostTag[T any](
	t *testing.T,
	tss *testServerSetup,
	tag any,
	expect T,
	statusCode int,
) {
	method := "POST"
	url := urlJoin(tss.uri, "producttag", "new")
	testRequestTag[T](t, tss, method, url, tag, expect, statusCode)
}

func testPutTag[T any](
	t *testing.T,
	tss *testServerSetup,
	url string,
	tag any,
	expect T,
	statusCode int,
) {
	method := "PUT"
	testRequestTag[T](t, tss, method, url, tag, expect, statusCode)
}

func testDeleteTag[T any](
	t *testing.T,
	tss *testServerSetup,
	tag *types.Tag,
	expect T,
	statusCode int,
) {
	method := "DELETE"
	url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d", tag.Id))
	testRequestTag[T](t, tss, method, url, tag, expect, statusCode)
}

func TestGetAllProductTags(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	expect := testProductTags
	lgc.On("GetAllProductTags").Return(expect, nil)

	url := urlJoin(tss.uri, "producttag")
	testGetTag[[]*types.Tag](t, tss, url, expect, http.StatusOK)
}

func TestGetAllProductTagsEmpty(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	expect := []*types.Tag{}
	lgc.On("GetAllProductTags").Return(expect, nil)

	url := urlJoin(tss.uri, "producttag")
	testGetTag[[]*types.Tag](t, tss, url, expect, http.StatusOK)
}

func TestGetAllProductTagsGetError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	lgc.On("GetAllProductTags").Return(nil, assert.AnError)

	url := urlJoin(tss.uri, "producttag")
	testGetTag[*apiError](t, tss, url, apiErrAnError, http.StatusInternalServerError)
}

func TestGetProductTagByID(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByID", expect.Id).Return(expect, nil)

		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d", expect.Id))
		testGetTag[*types.Tag](t, tss, url, expect, http.StatusOK)
	}
}

func TestGetProductTagByIDNotFound(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByID", expect.Id).Return(new(types.Tag), types.ErrIDNotFound)

		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d", expect.Id))
		testGetTag[*apiError](t, tss, url, apiErrNotFound, http.StatusNotFound)
	}
}

func TestGetProductTagByIDGetError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByID", expect.Id).Return(nil, assert.AnError)

		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d", expect.Id))
		testGetTag[*apiError](t, tss, url, apiErrAnError, http.StatusInternalServerError)
	}
}

func TestGetProductTagByIDWithParameter(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByID", expect.Id).Return(expect, nil)

		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d?searchby=id", expect.Id))
		testGetTag[*types.Tag](t, tss, url, expect, http.StatusOK)
	}
}

func TestGetProductTagByIDWithParameterNotFound(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByID", expect.Id).Return(new(types.Tag), types.ErrIDNotFound)

		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d?searchby=id", expect.Id))
		testGetTag[*apiError](t, tss, url, apiErrNotFound, http.StatusNotFound)
	}
}

func TestGetProductTagByIDWithParameterGetError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByID", expect.Id).Return(nil, assert.AnError)

		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d?searchby=id", expect.Id))
		testGetTag[*apiError](t, tss, url, apiErrAnError, http.StatusInternalServerError)
	}
}

func TestGetProductTagByName(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByName", expect.Name).Return(expect, nil)

		url := urlJoin(tss.uri, "producttag", expect.Name)
		testGetTag[*types.Tag](t, tss, url, expect, http.StatusOK)
	}
}

func TestGetProductTagByNameNotFound(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByName", expect.Name).Return(new(types.Tag), types.ErrIDNotFound)

		url := urlJoin(tss.uri, "producttag", expect.Name)
		testGetTag[*apiError](t, tss, url, apiErrNotFound, http.StatusNotFound)
	}
}

func TestGetProductTagByNameGetError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByName", expect.Name).Return(nil, assert.AnError)

		url := urlJoin(tss.uri, "producttag", expect.Name)
		testGetTag[*apiError](t, tss, url, apiErrAnError, http.StatusInternalServerError)
	}
}

func TestGetProductTagByNameWithParameter(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByName", expect.Name).Return(expect, nil)

		url := urlJoin(tss.uri, "producttag", expect.Name, "?", "searchby=name")
		testGetTag[*types.Tag](t, tss, url, expect, http.StatusOK)
	}
}

func TestGetProductTagByNameWithParameterNotFound(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByName", expect.Name).Return(new(types.Tag), types.ErrIDNotFound)

		url := urlJoin(tss.uri, "producttag", expect.Name, "?", "seachby=name")
		testGetTag[*apiError](t, tss, url, apiErrNotFound, http.StatusNotFound)
	}
}

func TestGetProductTagByNameWithParameterGetError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("GetProductTagByName", expect.Name).Return(nil, assert.AnError)

		url := urlJoin(tss.uri, "producttag", expect.Name, "?", "seachby=name")
		testGetTag[*apiError](t, tss, url, apiErrAnError, http.StatusInternalServerError)
	}
}

func TestProductTagsLikeName(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for search, expect := range map[string][]*types.Tag{
		"tag": testProductTags[:],
		"1":   testProductTags[0:1],
		"2":   testProductTags[1:],
		"3":   testProductTags[2:],
		"_":   {},
	} {
		lgc.On("GetProductTagsLikeName", search).Return(expect, nil)

		url := urlJoin(tss.uri, "producttag", "search", "?", "name="+search)
		testGetTag[[]*types.Tag](t, tss, url, expect, http.StatusOK)
	}
}

func TestProductTagsLikeNameGetError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for search, _ := range map[string][]*types.Tag{
		"tag": testProductTags[:],
		"1":   testProductTags[0:1],
		"2":   testProductTags[1:],
		"3":   testProductTags[2:],
		"_":   {},
	} {
		lgc.On("GetProductTagsLikeName", search).Return(nil, assert.AnError)

		url := urlJoin(tss.uri, "producttag", "search", "?", "name="+search)
		testGetTag[*apiError](t, tss, url, apiErrAnError, http.StatusInternalServerError)
	}
}

func TestPostProductTag(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("CreateProductTag", expect).Return(nil)

		testPostTag[*types.Tag](t, tss, expect, expect, http.StatusCreated)
	}
}

func TestPostProductTagJSONDecodeError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		data, err := json.Marshal(expect)
		require.NoError(t, err)

		errData := bytes.Replace(data, []byte(`"`), []byte(`42`), 3)
		err = json.Unmarshal(errData, new(types.Product))
		require.Error(t, err)

		expect := &apiError{
			Code:    http.StatusBadRequest,
			Message: err.Error(),
		}

		testPostProduct[*apiError](t, tss, errData, expect, http.StatusBadRequest)
	}
}

func TestPostProductTagCreateError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("CreateProductTag", expect).Return(assert.AnError)

		testPostTag[*apiError](t, tss, expect, apiErrAnError, http.StatusInternalServerError)
	}
}

func TestPutProductTag(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("UpdateProductTag", expect).Return(expect, nil)

		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d", expect.Id))
		testPutTag[*types.Tag](t, tss, url, expect, expect, http.StatusCreated)
	}
}

func TestPutProductTagJSONDecodeError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		data, err := json.Marshal(expect)
		require.NoError(t, err)

		errData := bytes.Replace(data, []byte(`"`), []byte(`42`), 3)
		err = json.Unmarshal(errData, new(types.Product))
		require.Error(t, err)

		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d", expect.Id))
		expect := &apiError{
			Code:    http.StatusBadRequest,
			Message: err.Error(),
		}

		testPutProduct[*apiError](t, tss, url, errData, expect, http.StatusBadRequest)
	}
}

func TestPutProductTagNotFound(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("UpdateProductTag", expect).Return(new(types.Tag), types.ErrIDNotFound)

		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d", expect.Id))
		testPutTag[*apiError](t, tss, url, expect, apiErrNotFound, http.StatusNotFound)
	}
}

func TestPutProductTagIDMismatch(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d", expect.Id+1))
		testPutTag[*apiError](t, tss, url, expect, apiErrIDMismatch, http.StatusBadRequest)
	}
}

func TestPutProductTagUpdateError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("UpdateProductTag", expect).Return(nil, assert.AnError)

		url := urlJoin(tss.uri, "producttag", fmt.Sprintf("%d", expect.Id))
		testPutTag[*apiError](t, tss, url, expect, apiErrAnError, http.StatusInternalServerError)
	}
}

func TestDeleteProductTag(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("DeleteProductTag", expect.Id).Return(expect, nil)

		testDeleteTag[*types.Tag](t, tss, expect, expect, http.StatusAccepted)
	}
}

func TestDeleteProductTagNotFound(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("DeleteProductTag", expect.Id).Return(new(types.Tag), types.ErrIDNotFound)

		testDeleteTag[*apiError](t, tss, expect, apiErrNotFound, http.StatusNotFound)
	}
}

func TestDeleteProductTagDeleteError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, expect := range testProductTags {
		lgc.On("DeleteProductTag", expect.Id).Return(nil, assert.AnError)

		testDeleteTag[*apiError](t, tss, expect, apiErrAnError, http.StatusInternalServerError)
	}
}
