/**
 * file: tests/api/v0/product_test.go
 * author: Theo Technicguy
 * license: Apache 2.0
 *
 * Products APIv0 endpoint tests
 */

package v0_test

import (
	"bytes"
	"encoding/json"
	"fmt"
	"net/http"
	"testing"

	"git.licolas.net/hoffman/server/types"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

var (
	testProducts = []*types.Product{
		{Id: 0x1, EAN: 0x0, Name: "Test1: P1 XN XE T", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0x2, EAN: 0x0, Name: "Test2: P1 XN XE T3", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0x3, EAN: 0x0, Name: "Test3: P1 XN XE T2", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0x4, EAN: 0x0, Name: "Test4: P1 XN XE T23", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0x5, EAN: 0x0, Name: "Test5: P1 XN XE T1", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0x6, EAN: 0x0, Name: "Test6: P1 XN XE T13", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0x7, EAN: 0x0, Name: "Test7: P1 XN XE T12", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0x8, EAN: 0x0, Name: "Test8: P1 XN XE T123", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0x9, EAN: 0x7b, Name: "Test9: P2 VN XE T", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0xa, EAN: 0x7c, Name: "Test10: P2 VN XE T3", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0xb, EAN: 0x7d, Name: "Test11: P2 VN XE T2", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0xc, EAN: 0x78, Name: "Test12: P2 VN XE T23", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0xd, EAN: 0x7f, Name: "Test13: P2 VN XE T1", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0xe, EAN: 0x80, Name: "Test14: P2 VN XE T13", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0xf, EAN: 0x81, Name: "Test15: P2 VN XE T12", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0x10, EAN: 0x82, Name: "Test16: P2 VN XE T123", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: nil},
		{Id: 0x11, EAN: 0x0, Name: "Test17: P3 XN VE T", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x12, EAN: 0x0, Name: "Test18: P3 XN VE T3", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x13, EAN: 0x0, Name: "Test19: P3 XN VE T2", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x14, EAN: 0x0, Name: "Test20: P3 XN VE T23", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x15, EAN: 0x0, Name: "Test21: P3 XN VE T1", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x16, EAN: 0x0, Name: "Test22: P3 XN VE T13", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x17, EAN: 0x0, Name: "Test23: P3 XN VE T12", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x18, EAN: 0x0, Name: "Test24: P3 XN VE T123", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x19, EAN: 0xea, Name: "Test25: P4 VN VE T", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x1a, EAN: 0xeb, Name: "Test26: P4 VN VE T3", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x1b, EAN: 0xec, Name: "Test27: P4 VN VE T2", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x1c, EAN: 0xed, Name: "Test28: P4 VN VE T23", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x1d, EAN: 0xee, Name: "Test29: P4 VN VE T1", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x1e, EAN: 0xef, Name: "Test30: P4 VN VE T13", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x1f, EAN: 0xf0, Name: "Test31: P4 VN VE T12", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
		{Id: 0x20, EAN: 0xf1, Name: "Test32: P4 VN VE T123", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: nil},
	}
)

func testRequestProduct[T any](
	t *testing.T,
	tss *testServerSetup,
	method string,
	url string,
	product any,
	expect T,
	statusCode int,
) {
	testRequestWithMarshal[T, *types.Product](t, tss, method, url, product, expect, statusCode)
}

func testGetProduct[T any](
	t *testing.T,
	tss *testServerSetup,
	url string,
	expect T,
	statusCode int,
) {
	method := "GET"
	testRequestProduct[T](t, tss, method, url, nil, expect, statusCode)
}

func testPostProduct[T any](
	t *testing.T,
	tss *testServerSetup,
	product any,
	expect T,
	statusCode int,
) {
	method := "POST"
	url := urlJoin(tss.uri, "product", "new")
	testRequestProduct[T](t, tss, method, url, product, expect, statusCode)
}

func testPutProduct[T any](
	t *testing.T,
	tss *testServerSetup,
	url string,
	product any,
	expect T,
	statusCode int,
) {
	method := "PUT"
	testRequestProduct[T](t, tss, method, url, product, expect, statusCode)
}

func testDeleteProduct[T any](
	t *testing.T,
	tss *testServerSetup,
	product *types.Product,
	expect T,
	statusCode int,
) {
	method := "DELETE"
	url := urlJoin(tss.uri, "product", fmt.Sprintf("%d", product.Id))
	testRequestProduct[T](t, tss, method, url, product, expect, statusCode)
}

func TestGetAllProducts(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	expect := testProducts
	lgc.On("GetAllProducts").Return(expect, nil)

	url := urlJoin(tss.uri, "product")
	testGetProduct[[]*types.Product](t, tss, url, expect, http.StatusOK)
}

func TestGetAllProductsEmpty(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	expect := []*types.Product{}
	lgc.On("GetAllProducts").Return(expect, nil)

	url := urlJoin(tss.uri, "product")
	testGetProduct[[]*types.Product](t, tss, url, expect, http.StatusOK)
}

func TestGetAllProductsError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	lgc.On("GetAllProducts").Return(nil, assert.AnError)

	url := urlJoin(tss.uri, "product")
	testGetProduct[*apiError](t, tss, url, apiErrAnError, http.StatusInternalServerError)
}

func TestGetProductByID(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("GetProductByID", product.Id).Return(product, nil)

		url := urlJoin(tss.uri, "product", fmt.Sprintf("%d", product.Id))
		testGetProduct[*types.Product](t, tss, url, product, http.StatusOK)
	}
}

func TestGetProductByIDNotFound(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("GetProductByID", product.Id).Return(new(types.Product), types.ErrIDNotFound)

		url := urlJoin(tss.uri, "product", fmt.Sprintf("%d", product.Id))
		testGetProduct[*apiError](t, tss, url, apiErrNotFound, http.StatusNotFound)
	}
}

func TestGetProductByIDError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("GetProductByID", product.Id).Return(new(types.Product), assert.AnError)

		url := urlJoin(tss.uri, "product", fmt.Sprintf("%d", product.Id))
		testGetProduct[*apiError](t, tss, url, apiErrAnError, http.StatusInternalServerError)
	}
}

func TestPostProduct(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("CreateProduct", product).Return(nil)

		testPostProduct[*types.Product](t, tss, product, product, http.StatusCreated)
	}
}

// TODO: GetProductsLike

func TestPostProductJSONDecodeError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		data, err := json.Marshal(product)
		require.NoError(t, err)

		errData := bytes.Replace(data, []byte(`"`), []byte(`42`), 3)
		err = json.Unmarshal(errData, new(types.Product))
		require.Error(t, err)

		expect := &apiError{
			Code:    http.StatusBadRequest,
			Message: err.Error(),
		}

		testPostProduct[*apiError](t, tss, errData, expect, http.StatusBadRequest)
	}
}

func TestPostProductCreateError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("CreateProduct", product).Return(assert.AnError)

		testPostProduct[*apiError](
			t,
			tss,
			product,
			apiErrAnError,
			http.StatusInternalServerError,
		)
	}
}

func TestPutProduct(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("UpdateProduct", product).Return(product, nil)

		url := urlJoin(tss.uri, "product", fmt.Sprintf("%d", product.Id))
		testPutProduct[*types.Product](t, tss, url, product, product, http.StatusCreated)
	}
}

func TestPutProductJSONDecodeError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		data, err := json.Marshal(product)
		require.NoError(t, err)

		errData := bytes.Replace(data, []byte(`"`), []byte(`42`), 3)
		err = json.Unmarshal(errData, new(types.Product))
		require.Error(t, err)

		expect := &apiError{
			Code:    http.StatusBadRequest,
			Message: err.Error(),
		}

		url := urlJoin(tss.uri, "product", fmt.Sprintf("%d", product.Id))
		testPutProduct[*apiError](t, tss, url, errData, expect, http.StatusBadRequest)
	}
}

func TestPutProductIDMismatch(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("UpdateProduct", product).Return(product, nil)

		url := urlJoin(tss.uri, "product", fmt.Sprintf("%d", product.Id+1))
		testPutProduct[*apiError](
			t,
			tss,
			url,
			product,
			apiErrIDMismatch,
			http.StatusBadRequest,
		)
	}
}

func TestPutProductNotFound(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("UpdateProduct", product).Return(new(types.Product), types.ErrIDNotFound)

		url := urlJoin(tss.uri, "product", fmt.Sprintf("%d", product.Id))
		testPutProduct[*apiError](t, tss, url, product, apiErrNotFound, http.StatusNotFound)
	}
}

func TestPutProductUpdateError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("UpdateProduct", product).Return(nil, assert.AnError)

		url := urlJoin(tss.uri, "product", fmt.Sprintf("%d", product.Id))
		testPutProduct[*apiError](
			t,
			tss,
			url,
			product,
			apiErrAnError,
			http.StatusInternalServerError,
		)
	}
}

func TestDeleteProduct(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("DeleteProduct", product.Id).Return(product, nil)

		testDeleteProduct[*types.Product](t, tss, product, product, http.StatusAccepted)
	}
}

func TestDeleteProductNotFound(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("DeleteProduct", product.Id).Return(new(types.Product), types.ErrIDNotFound)

		testDeleteProduct[*apiError](t, tss, product, apiErrNotFound, http.StatusNotFound)
	}
}

func TestDeleteProductDeleteError(t *testing.T) {
	lgc := new(mockedLogic)
	tss := setupTest(t, lgc)
	defer teardownTest(tss)

	for _, product := range testProducts {
		lgc.On("DeleteProduct", product.Id).Return(nil, assert.AnError)

		testDeleteProduct[*apiError](
			t,
			tss,
			product,
			apiErrAnError,
			http.StatusInternalServerError,
		)
	}
}
