/**
 * file: tests/api/v0/utils_test.go
 * author: Theo Technicguy
 * license: Apache 2.0
 *
 * API tests utility functions
 */

package v0_test

import (
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"io"
	"math/rand"
	"net/http"
	"strings"
	"testing"
	"time"

	"git.licolas.net/hoffman/server/api"
	apiv0 "git.licolas.net/hoffman/server/api/v0"
	"git.licolas.net/hoffman/server/types"
	"github.com/gorilla/mux"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

const (
	serverAddr = "[::1]:%d"         // %d will be filled with a random port
	serverURI  = "http://%s/api/v0" // %s will be filled with the port-filled server address

	// --- Test Error Messages ---
	// Test teardown
	tlmServerCloseTooEarly = "http api server closed prematurely"
	// Test request
	tlmFailRequest = "failed to %s %s" // req.Method, req.URL
)

var (
	// Common errors
	apiErrAnError = &apiError{
		Code:    http.StatusInternalServerError,
		Message: assert.AnError.Error(),
	}
	apiErrNotFound = &apiError{
		Code:    http.StatusNotFound,
		Message: types.ErrIDNotFound.Error(),
	}
	apiErrIDMismatch = &apiError{
		Code:    http.StatusBadRequest,
		Message: apiv0.ErrMismatchingIds.Error(),
	}
)

type testServerSetup struct {
	server   *http.Server
	client   *http.Client
	uri      string
	mayClose bool
}

func (tss *testServerSetup) Shutdown() {
	ctx, cancel := context.WithTimeout(context.Background(), 250*time.Millisecond)
	defer cancel()

	tss.server.Shutdown(ctx)
}

type apiError struct {
	Code    int    `json:"code"`
	Message string `json:"message"`
}

func urlJoin(parts ...string) string {
	var i int
	for i = 0; i < len(parts); i++ {
		if parts[i] == "?" {
			break
		}
	}

	url := strings.Join(parts[:i], "/")
	var params string
	if i < len(parts) {
		params = "?" + strings.Join(parts[i+1:], "&")
	} else {
		params = ""
	}

	return url + params
}

func setupTest(t *testing.T, lgc types.LogicProvider) *testServerSetup {
	router := mux.NewRouter()
	apiRouter := router.PathPrefix("/api").Subrouter()
	api.SetLogicProvider(lgc)
	api.SetRoutes(apiRouter)

	port := 0
	for port < 0x800 {
		port = rand.Int() % 0xffff
	}
	addr := fmt.Sprintf(serverAddr, port)

	tss := &testServerSetup{
		server: &http.Server{
			Addr:    addr,
			Handler: router,
		},
		client: &http.Client{
			Timeout: time.Second,
		},
		mayClose: false,
		uri:      fmt.Sprintf(serverURI, addr),
	}

	go func() {
		err := tss.server.ListenAndServe()
		if !assert.True(t, tss.mayClose, tlmServerCloseTooEarly) {
			assert.NoError(t, err)
		}
	}()

	time.Sleep(100 * time.Millisecond)

	return tss
}

func teardownTest(tss *testServerSetup) {
	tss.mayClose = true
	tss.Shutdown()
}

func testHeader(t *testing.T, resp *http.Response) {
	contentType := resp.Header.Get("Content-Type")
	assert.Equal(t, "application/json; charset=utf-8", contentType)
}

// testRequest adds the json content type and executes the request.
// Then it checks the status code and parses the json from the
// server using the generic type, which it compares to the expected
// value.
func testRequest[T any](
	t *testing.T,
	tss *testServerSetup,
	req *http.Request,
	expect T,
	statusCode int,
) {
	require.NotNil(t, req)
	req.Header.Set("Content-Type", "application/json")
	resp, err := tss.client.Do(req)
	require.NoErrorf(t, err, tlmFailRequest, req.Method, req.URL)

	defer resp.Body.Close()

	testHeader(t, resp)

	body, err := io.ReadAll(resp.Body)
	require.NoError(t, err)
	assert.Equal(t, statusCode, resp.StatusCode)

	var actual T
	err = json.Unmarshal(body, &actual)
	require.NoError(t, err)
	assert.Equal(t, expect, actual)
}

// testRequestWithMarshal marshalls data if it matches V.
// It then passes everything to testRequest
func testRequestWithMarshal[T, V any](
	t *testing.T,
	tss *testServerSetup,
	method string,
	url string,
	data any,
	expect T,
	statusCode int,
) {
	var buf io.Reader
	switch v := data.(type) {
	case V:
		data, err := json.Marshal(v)
		require.NoError(t, err)
		buf = bytes.NewBuffer(data)
	case []byte:
		buf = bytes.NewBuffer(v)
	case nil:
		buf = nil
	default:
		assert.FailNow(t, ErrConversion.Error())
	}

	req, err := http.NewRequest(method, url, buf)
	require.NoError(t, err)

	testRequest[T](t, tss, req, expect, statusCode)
}
