/**
 * file: tests/database/productTag_test.go
 * author: Theo Technicguy
 * license: Apache 2.0
 *
 * Database product connector tests
 */

package database_test

import (
	"database/sql"
	_ "embed"
	"testing"

	"git.licolas.net/hoffman/server/database"
	"git.licolas.net/hoffman/server/types"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func setupTestProductTag() []*types.Tag {
	return []*types.Tag{
		{Id: 1, Name: "tag1", Color: 0},
		{Id: 2, Name: "tag2", Color: 123456},
		{Id: 3, Name: "tag3", Color: 16581375},
	}
}

func testCreateProductTagTable(t *testing.T, db *sql.DB) {
	_, err := db.Exec(`CREATE TABLE ProductTag (
		TagID INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
		Name VARCHAR(64) UNIQUE NOT NULL,
		Color INTEGER
	);`)
	require.NoError(t, err)
}

func testInsertProductTag(t *testing.T, db *sql.DB, tag *types.Tag) {
	query := "INSERT INTO ProductTag (TagID, Name, Color) VALUES (?, ?, ?)"
	args := []any{tag.Id, tag.Name, tag.Color}
	testExec(t, db, query, args...)
}

func testSelectProductTag(
	t *testing.T,
	db *sql.DB,
	query string,
	args ...any,
) []*types.Tag {
	scanner := func(rows *sql.Rows) (tag *types.Tag, err error) {
		tag = new(types.Tag)
		err = rows.Scan(&tag.Id, &tag.Name, &tag.Color)
		return
	}
	return testSelect[*types.Tag](t, db, scanner, query, args...)
}

func testSelectProductTagByID(
	t *testing.T,
	db *sql.DB,
	id uint,
) *types.Tag {
	query := "SELECT TagID, Name, Color FROM ProductTag WHERE TagID = ?"
	tags := testSelectProductTag(t, db, query, id)
	require.LessOrEqual(t, len(tags), 1)

	if len(tags) == 1 {
		return tags[0]
	} else {
		return new(types.Tag)
	}
}

func testSelectAllProductTag(t *testing.T, db *sql.DB, expect []*types.Tag) []*types.Tag {
	query := "SELECT TagID, Name, Color FROM ProductTag"
	return testSelectProductTag(t, db, query)
}

func TestInsertProductTag(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	for _, tag := range setupTestProductTag() {
		err := dsrc.InsertProductTag(tag)
		require.NoError(t, err)
		actual := testSelectProductTagByID(t, db, tag.Id)
		assert.Equal(t, tag, actual)
	}
}

func TestInsertProductTagError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	for _, tag := range setupTestProductTag() {
		err := dsrc.InsertProductTag(tag)
		require.Error(t, err)
	}
}

func TestSelectAllProductTags(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}

	actual, err := dsrc.SelectAllProductTags()
	require.NoError(t, err)
	assert.ElementsMatch(t, tags, actual)
}

func TestSelectAllProductTagsEmpty(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := []*types.Tag{}

	actual, err := dsrc.SelectAllProductTags()
	require.NoError(t, err)
	assert.ElementsMatch(t, tags, actual)
}

func TestSelectAllProductTagsError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	actual, err := dsrc.SelectAllProductTags()
	require.Error(t, err)
	assert.Nil(t, actual)
}

func TestSelectProductTagByID(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}

	for _, tag := range tags {
		actual, err := dsrc.SelectProductTagById(tag.Id)
		require.NoError(t, err)

		assert.Equal(t, tag, actual)
	}
}

func TestSelectProductTagByIDNotFound(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()

	for _, tag := range tags {
		actual, err := dsrc.SelectProductTagById(tag.Id)
		require.NoError(t, err)

		assert.Equal(t, new(types.Tag), actual)
	}
}

func TestSelectProductTagByIDError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()

	for _, tag := range tags {
		actual, err := dsrc.SelectProductTagById(tag.Id)
		require.Error(t, err)

		assert.Nil(t, actual)
	}
}

func TestSelectProductTagByName(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}

	for _, tag := range tags {
		actual, err := dsrc.SelectProductTagByName(tag.Name)
		require.NoError(t, err)

		assert.Equal(t, tag, actual)
	}
}

func TestSelectProductTagByNameNotFound(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()

	for _, tag := range tags {
		actual, err := dsrc.SelectProductTagByName(tag.Name)
		require.NoError(t, err)

		assert.Equal(t, new(types.Tag), actual)
	}
}

func TestSelectProductTagByNameError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()

	for _, tag := range tags {
		actual, err := dsrc.SelectProductTagByName(tag.Name)
		require.Error(t, err)

		assert.Nil(t, actual)
	}
}

func TestUpdateProductTag(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}

	for _, initial := range tags {
		for _, update := range tags {
			tag := new(types.Tag)
			tag.Id = initial.Id
			tag.Name = update.Name
			tag.Color = update.Color

			updated, err := dsrc.UpdateProductTag(tag)
			actual := testSelectProductTagByID(t, db, tag.Id)

			// Names are supposed to be unique
			// Updating them should produce an error
			// When the update throws an error, the old tag
			// should still be in the database, unchanged.
			if update.Name == initial.Name {
				assert.NoError(t, err)
				assert.Equal(t, tag, actual)
				assert.Equal(t, tag, updated)
			} else {
				assert.Error(t, err)
				assert.NotEqual(t, tag, actual)
				assert.Equal(t, initial, actual)
			}
		}
	}
}

func TestUpdateProductTagNotFound(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}

	for _, initial := range tags {
		tag := new(types.Tag)
		tag.Id = 0
		tag.Name = initial.Name
		tag.Color = initial.Color

		updated, err := dsrc.UpdateProductTag(tag)
		actual := testSelectProductTagByID(t, db, initial.Id)
		assert.NoError(t, err)
		assert.Equal(t, initial, actual)
		assert.Equal(t, new(types.Tag), updated)
	}
}

func TestUpdateProductTagError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, initial := range tags {
		for _, updated := range tags {
			tag := new(types.Tag)
			tag.Id = initial.Id
			tag.Name = updated.Name
			tag.Color = updated.Color

			updated, err := dsrc.UpdateProductTag(tag)
			assert.Error(t, err)
			assert.Nil(t, updated)
		}
	}
}

func TestDeleteProductTag(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	testCreateProductTable(t, db)
	testCreateProductTaggingTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
		actual, err := dsrc.DeleteProductTag(tag.Id)
		require.NoError(t, err)
		assert.Equal(t, tag, actual)
		testSelectAllProductTag(t, db, nil)
	}
}

func TestDeleteProductTagNotFound(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
		actual, err := dsrc.DeleteProductTag(0)
		assert.Error(t, err)
		assert.Nil(t, actual)

		notDel := testSelectProductTagByID(t, db, tag.Id)
		assert.Equal(t, tag, notDel)
	}
}

func TestDeleteProductTagError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		actual, err := dsrc.DeleteProductTag(tag.Id)
		assert.Error(t, err)
		assert.Nil(t, actual)
	}
}
