/**
 * file: tests/database/product_test.go
 * author: Theo Technicguy
 * license: Apache 2.0
 *
 * Database product connector tests
 */

package database_test

import (
	"database/sql"
	"fmt"
	"testing"

	"git.licolas.net/hoffman/server/database"
	"git.licolas.net/hoffman/server/types"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

// The data source is supposed to change the IDs from new entries. This will break
// consecutive tests.
func setupTestProducts() []*types.Product {
	tags := setupTestProductTag()
	combs := (0b1 << len(tags)) - 1

	testProductTTLs := []*types.TTL{
		{Normal: 0, Fridge: 0, Freezer: 0},
		{Normal: 0, Fridge: 14, Freezer: 180},
		{Normal: 180, Fridge: 7, Freezer: 980},
	}

	testProductEANs := []uint{0, 5400141301571}

	var id uint = 1
	testProducts := []*types.Product{}
	for _, ean := range testProductEANs {
		for _, ttl := range testProductTTLs {
			for tag := 0; tag < combs; tag++ {
				p := new(types.Product)
				p.Id = id
				p.EAN = ean
				if p.EAN > 0 {
					p.EAN += id
				}
				p.Name = fmt.Sprintf("Test%d", id)
				p.TTL = *ttl
				p.Tags = []types.Tag{}

				for tdx, tg := range tags {
					if tag&(0b1<<tdx) != 0 {
						p.Tags = append(p.Tags, *tg)
					}
				}

				testProducts = append(testProducts, p)

				id += 1
			}
		}
	}

	return testProducts
}

func testCreateProductTable(t *testing.T, db *sql.DB) {
	_, err := db.Exec(`CREATE TABLE Product (
		ProductID INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
		EAN INTEGER UNIQUE,
		Name VARCHAR(64) NOT NULL,
		TTL INTEGER,
		TTLFridge INTEGER,
		TTLFreezer INTEGER
	)`)
	require.NoError(t, err)
}

func testInsertProduct(t *testing.T, db *sql.DB, p *types.Product) {
	query := `INSERT INTO Product 
		(ProductID, EAN, Name, TTL, TTLFridge, TTLFreezer)
	VALUES
		(?, ?, ?, ?, ?, ?);`
	var ean any
	if p.EAN == 0 {
		ean = nil
	} else {
		ean = p.EAN
	}
	args := []any{p.Id, ean, p.Name, p.TTL.Normal, p.TTL.Fridge, p.TTL.Freezer}
	testExec(t, db, query, args...)

	for _, tag := range p.Tags {
		testInsertProductTagging(t, db, p.Id, tag.Id)
	}
}

func testSelectProduct(
	t *testing.T,
	db *sql.DB,
	query string,
	args ...any,
) []*types.Product {
	scanner := func(rows *sql.Rows) (p *types.Product, err error) {
		p = new(types.Product)
		p.TTL = types.TTL{}
		var ean any
		err = rows.Scan(&p.Id, &ean, &p.Name, &p.TTL.Normal, &p.TTL.Fridge, &p.TTL.Freezer)

		if ean == nil {
			p.EAN = 0
		} else {
			p.EAN = uint(ean.(int64))
		}

		return p, err
	}
	products := testSelect[*types.Product](t, db, scanner, query, args...)

	for pdx, p := range products {
		products[pdx].Tags = testSelectTaggingsByProductID(t, db, p.Id)
	}

	return products
}

func testSelectProductByID(
	t *testing.T,
	db *sql.DB,
	id uint,
) *types.Product {
	query := `SELECT ProductID, EAN, Name, TTL, TTLFridge, TTLFreezer
	FROM Product
	WHERE ProductID = ?;`
	products := testSelectProduct(t, db, query, id)
	require.LessOrEqual(t, len(products), 1)

	if len(products) == 1 {
		return products[0]
	} else {
		return new(types.Product)
	}
}

func testSelectAllProducts(
	t *testing.T,
	db *sql.DB,
) []*types.Product {
	query := `SELECT ProductID, EAN, Name, TTL, TTLFridge, TTLFreezer
	FROM Product;`
	products := testSelectProduct(t, db, query)

	for _, product := range products {
		product.Tags = testSelectTaggingsByProductID(t, db, product.Id)
	}

	return products
}

func TestInsertProduct(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTable(t, db)
	testCreateProductTagTable(t, db)
	testCreateProductTaggingTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	products := setupTestProducts()
	expectProducts := []*types.Product{}
	for _, product := range products {
		err := dsrc.InsertProduct(product)
		require.NoError(t, err)

		actual := testSelectProductByID(t, db, product.Id)
		assert.Equal(t, product, actual)

		expectProducts = append(expectProducts, product)
		actualProducts := testSelectAllProducts(t, db)
		assert.ElementsMatch(t, expectProducts, actualProducts)
	}
}

func TestInsertProductError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	products := setupTestProducts()
	for _, product := range products {
		err := dsrc.InsertProduct(product)
		assert.Error(t, err)
	}
}

func TestSelectAllProducts(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTable(t, db)
	testCreateProductTagTable(t, db)
	testCreateProductTaggingTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}
	products := setupTestProducts()
	for _, product := range products {
		testInsertProduct(t, db, product)
	}

	actual, err := dsrc.SelectAllProducts()
	require.NoError(t, err)
	assert.ElementsMatch(t, products, actual)
}

func TestSelectAllProductsError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	actual, err := dsrc.SelectAllProducts()
	assert.Error(t, err)
	assert.Nil(t, actual)
}

func TestSelectProductByID(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTable(t, db)
	testCreateProductTagTable(t, db)
	testCreateProductTaggingTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}
	products := setupTestProducts()
	for _, product := range products {
		testInsertProduct(t, db, product)
	}

	for _, expect := range products {
		actual, err := dsrc.SelectProductById(expect.Id)
		require.NoError(t, err)
		assert.Equal(t, expect, actual)
	}
}

func TestSelectProductByIDNotFound(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTable(t, db)
	testCreateProductTagTable(t, db)
	testCreateProductTaggingTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	products := setupTestProducts()
	for _, p := range products {
		actual, err := dsrc.SelectProductById(p.Id)
		require.NoError(t, err)
		assert.Equal(t, new(types.Product), actual)
	}
}

func TestSelectProductByIDError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	products := setupTestProducts()
	for _, p := range products {
		actual, err := dsrc.SelectProductById(p.Id)
		require.Error(t, err)
		assert.Nil(t, actual)
	}
}

// TODO: TestSelectProductLike...

func TestUpdateProduct(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTable(t, db)
	testCreateProductTagTable(t, db)
	testCreateProductTaggingTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	assert := assert.New(t)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}
	products := setupTestProducts()
	for _, product := range products {
		testInsertProduct(t, db, product)
	}

	for _, initial := range products {
		for _, update := range products {
			product := new(types.Product)
			product.Id = initial.Id
			product.EAN = update.EAN
			product.Name = update.Name
			product.TTL.Normal = update.TTL.Normal
			product.TTL.Fridge = update.TTL.Fridge
			product.TTL.Freezer = update.TTL.Freezer
			product.Tags = update.Tags
			t.Logf("Updating (initial) %d (product) %d to (update) %d", initial.Id, product.Id, update.Id)

			updated, err := dsrc.UpdateProduct(product)
			actual := testSelectProductByID(t, db, product.Id)

			// Check that the product didn't change
			assert.Equal(initial.Id, product.Id)
			assert.Equal(update.EAN, product.EAN)
			assert.Equal(update.Name, product.Name)
			assert.Equal(update.TTL, product.TTL)
			assert.ElementsMatch(update.Tags, product.Tags)

			// EANs are supposed to be unique
			// Updating them should produce an error
			// When the update throws an error, the old product
			// should still be in the database, unchanged.
			if product.EAN == 0 || initial.EAN == update.EAN {
				assert.NoError(err)

				// Check the returned update
				assert.Equal(initial.Id, updated.Id)
				assert.Equal(update.EAN, updated.EAN)
				assert.Equal(update.Name, updated.Name)
				assert.Equal(update.TTL, updated.TTL)
				assert.ElementsMatch(update.Tags, updated.Tags)

				// Check the actual update
				assert.Equal(initial.Id, actual.Id)
				assert.Equal(update.EAN, actual.EAN)
				assert.Equal(update.Name, actual.Name)
				assert.Equal(update.TTL, actual.TTL)
				assert.ElementsMatch(update.Tags, actual.Tags)
			} else {
				assert.Error(err)
				assert.Nil(updated)
			}
		}
	}
}

func TestUpdateProductNoFound(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTable(t, db)
	testCreateProductTagTable(t, db)
	testCreateProductTaggingTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}
	products := setupTestProducts()
	for _, product := range products {
		testInsertProduct(t, db, product)
	}

	for _, initial := range products {
		product := new(types.Product)
		product.Id = 0
		product.EAN = initial.EAN
		product.Name = initial.Name
		product.TTL.Normal = initial.TTL.Normal
		product.TTL.Fridge = initial.TTL.Fridge
		product.TTL.Freezer = initial.TTL.Freezer

		updated, err := dsrc.UpdateProduct(product)
		actual := testSelectProductByID(t, db, initial.Id)
		assert.NoError(t, err)
		assert.Equal(t, new(types.Product), updated)
		assert.Equal(t, initial, actual)
	}
}

func TestUpdateProductError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	products := setupTestProducts()
	for _, initial := range products {
		for _, update := range products {
			product := new(types.Product)
			product.Id = initial.Id
			product.EAN = update.EAN
			product.Name = update.Name
			product.TTL.Normal = update.TTL.Normal
			product.TTL.Fridge = update.TTL.Fridge
			product.TTL.Freezer = update.TTL.Freezer

			updated, err := dsrc.UpdateProduct(product)
			assert.Error(t, err)
			assert.Nil(t, updated)
		}
	}
}

func TestDeleteProduct(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	testCreateProductTable(t, db)
	testCreateProductTaggingTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}
	products := setupTestProducts()
	for _, product := range products {
		testInsertProduct(t, db, product)
	}

	expectProducts := setupTestProducts()
	for _, product := range products {
		actual, err := dsrc.DeleteProduct(product.Id)
		require.NoError(t, err)
		assert.Equal(t, product, actual)

		expectProducts = expectProducts[1:]
		actualProducts := testSelectAllProducts(t, db)
		assert.ElementsMatch(t, expectProducts, actualProducts)
	}
}

func TestDeleteProductNotFound(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateProductTagTable(t, db)
	testCreateProductTable(t, db)
	testCreateProductTaggingTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	tags := setupTestProductTag()
	for _, tag := range tags {
		testInsertProductTag(t, db, tag)
	}
	products := setupTestProducts()
	for _, product := range products {
		testInsertProduct(t, db, product)
	}

	expectProducts := setupTestProducts()
	for _, expect := range products {
		actual, err := dsrc.DeleteProduct(0)
		assert.NoError(t, err)
		assert.Equal(t, new(types.Product), actual)

		notDel := testSelectProductByID(t, db, expect.Id)
		assert.Equal(t, expect, notDel)

		actualProducts := testSelectAllProducts(t, db)
		assert.ElementsMatch(t, expectProducts, actualProducts)
	}
}

func TestDeleteProductError(t *testing.T) {
	fp := tmpPath()
	db := setupTest(t, fp)
	testCreateTempTable(t, db)
	dsrc, err := database.Connect(fp)
	require.NoError(t, err)
	defer teardownTest(t, dsrc, db, fp)

	products := setupTestProducts()
	for _, expect := range products {
		actual, err := dsrc.DeleteProduct(expect.Id)
		assert.Error(t, err)
		assert.Nil(t, actual)
	}
}
