/**
 * file: database/utils_test.go
 * author: Theo Technicguy
 * license: Apache 2.0
 *
 * Miscallenious database testing utilities
 */

package database_test

import (
	"database/sql"
	"fmt"
	"math/rand"
	"os"
	"testing"

	"git.licolas.net/hoffman/server/database"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

const (
	sqlTableStructure = "PRAGMA table_info(?):"
)

type column struct {
	Cid        uint
	Name       string
	Type       string
	NotNull    bool
	Default    any
	PrimaryKey uint
}

func tmpPath() string {
	return fmt.Sprintf("tmp-%d.db", rand.Int())
}

func setupTest(t *testing.T, fp string) *sql.DB {
	db, err := sql.Open("sqlite3", fp)
	require.NoError(t, err)
	return db
}

func teardownTest(t *testing.T, dsrc *database.R, tdb *sql.DB, fp string) {
	err := tdb.Close()
	assert.NoError(t, err)

	err = dsrc.Disconnect()
	assert.NoError(t, err)

	err = os.Remove(fp)
	assert.NoError(t, err)
}

func testCreateTempTable(t *testing.T, db *sql.DB) {
	query := "CREATE TABLE test (id INTEGER);"
	testExec(t, db, query)
}

func testTableStructure(
	t *testing.T,
	table string,
	expect []*column,
	tdb *sql.DB,
) {
	rows, err := tdb.Query(sqlTableStructure, table)
	require.NoError(t, err)

	cols := []*column{}
	for rows.Next() {
		var col *column
		err := rows.Scan(
			&col.Cid,
			&col.Name,
			&col.Type,
			&col.NotNull,
			&col.Default,
			&col.PrimaryKey,
		)
		require.NoError(t, err)

		cols = append(cols, col)
	}

	assert.ElementsMatch(t, cols, expect)
}

func testExec(t *testing.T, db *sql.DB, query string, args ...any) {
	_, err := db.Exec(query, args...)
	require.NoError(t, err)
}

func testSelect[T any](
	t *testing.T,
	db *sql.DB,
	scanner func(*sql.Rows) (T, error),
	query string,
	args ...any,
) []T {
	rows, err := db.Query(query, args...)
	require.NoError(t, err)

	actual := []T{}
	for rows.Next() {
		item, err := scanner(rows)
		require.NoError(t, err)
		actual = append(actual, item)
	}

	return actual
}
