/**
 * file: tests/logic/productTag_test.go
 * author: Theo Technicguy
 * license: Apache 2.0
 *
 * Logic product tests
 */

package logic_test

import (
	"testing"

	"git.licolas.net/hoffman/server/logic"
	"git.licolas.net/hoffman/server/types"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

var (
	tlmIncorrectProductTag  = "incorrect product tag received"
	tlmIncorrectProductTags = "incorrect product tags received"

	testProductTags = []*types.Tag{
		{Id: 0x1, Name: "tag1", Color: 0},
		{Id: 0x2, Name: "tag2", Color: 123456},
		{Id: 0x3, Name: "tag3", Color: 16581375},
	}
)

func TestCreateProductTag(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("InsertProductTag", tag).Return(nil)
		err := lgc.CreateProductTag(tag)
		assert.NoError(t, err)
	}
}

func TestCreateProductTagInsertError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("InsertProductTag", tag).Return(assert.AnError)
		err := lgc.CreateProductTag(tag)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
	}
}

func TestGetAllProductTags(t *testing.T) {
	mock := new(mockedDataSource)
	mock.On("SelectAllProductTags").Return(testProductTags, nil)
	lgc := logic.UseDatasource(mock)

	tags, err := lgc.GetAllProductTags()
	require.NoError(t, err)
	assert.ElementsMatch(t, tags, testProductTags, tlmIncorrectProductTags)
}

func TestGetAllProductTagsEmpty(t *testing.T) {
	mock := new(mockedDataSource)
	mock.On("SelectAllProductTags").Return([]*types.Tag{}, nil)
	lgc := logic.UseDatasource(mock)

	tags, err := lgc.GetAllProductTags()
	require.NoError(t, err)
	assert.ElementsMatch(t, tags, []*types.Tag{}, tlmIncorrectProductTag)
}

func TestGetAllProductTagsSelectError(t *testing.T) {
	mock := new(mockedDataSource)
	mock.On("SelectAllProductTags").Return(nil, assert.AnError)
	lgc := logic.UseDatasource(mock)

	tags, err := lgc.GetAllProductTags()
	require.Error(t, err)
	assert.Nil(t, tags)
}

func TestGetProductTagByID(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagById", tag.Id).Return(tag, nil)
		actual, err := lgc.GetProductTagByID(tag.Id)
		require.NoError(t, err)
		assert.Equal(t, actual, tag, tlmIncorrectProductTag)
	}
}

func TestGetProductTagByIDEmpty(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	mock.On("SelectProductTagById", uint(0)).Return(&types.Tag{}, nil)
	actual, err := lgc.GetProductTagByID(0)
	require.Error(t, err)
	assert.ErrorIs(t, err, types.ErrIDNotFound)
	assert.Equal(t, &types.Tag{}, actual, tlmIncorrectProductTag)
}

func TestGetProductTagByIDSelectError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagById", tag.Id).Return(nil, assert.AnError)
		actual, err := lgc.GetProductTagByID(tag.Id)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
		assert.Nil(t, actual, tlmIncorrectProductTag)
	}
}

func TestGetProductTagByName(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagByName", tag.Name).Return(tag, nil)
		actual, err := lgc.GetProductTagByName(tag.Name)
		require.NoError(t, err)
		assert.Equal(t, actual, tag, tlmIncorrectProductTag)
	}
}

func TestGetProductTagByNameEmpty(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	mock.On("SelectProductTagByName", "").Return(&types.Tag{}, nil)
	actual, err := lgc.GetProductTagByName("")
	require.Error(t, err)
	assert.ErrorIs(t, err, types.ErrIDNotFound)
	assert.Equal(t, &types.Tag{}, actual, tlmIncorrectProductTag)
}

func TestGetProductTagByNameSelectError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagByName", tag.Name).Return(nil, assert.AnError)
		actual, err := lgc.GetProductTagByName(tag.Name)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
		assert.Nil(t, actual, tlmIncorrectProductTag)
	}
}

func TestGetProductTagLikeName(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for term, result := range map[string][]*types.Tag{
		"tag": testProductTags[:],
		"1":   testProductTags[:1],
		"2":   testProductTags[1:2],
		"3":   testProductTags[2:],
		"_":   {},
	} {
		mock.On("SelectProductTagsLikeName", term).Return(result, nil)
		actual, err := lgc.GetProductTagsLikeName(term)
		require.NoError(t, err)
		assert.ElementsMatch(t, actual, result, tlmIncorrectProductTag)
	}
}

func TestGetProductTagLikeNameSelectError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)
	mock.On("SelectProductTagsLikeName", "tag").Return(nil, assert.AnError)
	actual, err := lgc.GetProductTagsLikeName("tag")
	require.Error(t, err)
	assert.ErrorIs(t, err, assert.AnError)
	assert.Nil(t, actual, tlmIncorrectProductTag)
}

func TestUpdateProductTag(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagById", tag.Id).Return(tag, nil)
		mock.On("UpdateProductTag", tag).Return(tag, nil)
		actual, err := lgc.UpdateProductTag(tag)
		assert.NoError(t, err)
		assert.Equal(t, tag, actual)
	}
}

func TestUpdateProductTagNotFound(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagById", tag.Id).Return(&types.Tag{}, nil)
		actual, err := lgc.UpdateProductTag(tag)
		require.Error(t, err)
		assert.ErrorIs(t, err, types.ErrIDNotFound)
		assert.Equal(t, new(types.Tag), actual)
	}
}

func TestUpdateProductTagSelectError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagById", tag.Id).Return(nil, assert.AnError)
		actual, err := lgc.UpdateProductTag(tag)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
		assert.Nil(t, actual)
	}
}

func TestUpdateProductTagUpdateError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagById", tag.Id).Return(tag, nil)
		mock.On("UpdateProductTag", tag).Return(nil, assert.AnError)
		actual, err := lgc.UpdateProductTag(tag)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
		assert.Nil(t, actual)
	}
}

func TestDeleteProductTag(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagById", tag.Id).Return(tag, nil)
		mock.On("DeleteProductTag", tag.Id).Return(tag, nil)
		actual, err := lgc.DeleteProductTag(tag.Id)
		require.NoError(t, err)
		assert.Equal(t, tag, actual, tlmIncorrectProductTag)
	}
}

func TestDeleteProductTagNotFound(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagById", tag.Id).Return(&types.Tag{}, nil)
		actual, err := lgc.DeleteProductTag(tag.Id)
		require.Error(t, err)
		assert.ErrorIs(t, err, types.ErrIDNotFound)
		assert.Nil(t, actual, tlmIncorrectProductTag)
	}
}

func TestDeleteProductTagSelectError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagById", tag.Id).Return(nil, assert.AnError)
		actual, err := lgc.DeleteProductTag(tag.Id)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
		assert.Nil(t, actual, tlmIncorrectProductTag)
	}
}

func TestDeleteProductTagError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, tag := range testProductTags {
		mock.On("SelectProductTagById", tag.Id).Return(tag, nil)
		mock.On("DeleteProductTag", tag.Id).Return(nil, assert.AnError)
		actual, err := lgc.DeleteProductTag(tag.Id)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
		assert.Nil(t, actual, tlmIncorrectProductTag)
	}
}
