/**
 * file: tests/logic/product_test.go
 * author: Theo Technicguy
 * license: Apache 2.0
 *
 * Logic product tests
 */

package logic_test

import (
	"testing"

	"git.licolas.net/hoffman/server/logic"
	"git.licolas.net/hoffman/server/types"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

var (
	// Test Log Messages
	tlmIncorrectProduct  = "incorrect product received"
	tlmIncorrectProducts = "incorrect products received"

	testProducts = []*types.Product{
		{Id: 0x1, EAN: 0x0, Name: "Test1: P1 XN XE T", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{}},
		{Id: 0x2, EAN: 0x0, Name: "Test2: P1 XN XE T3", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x3, EAN: 0x0, Name: "Test3: P1 XN XE T2", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x2, Name: "tag2", Color: 0x1e240}}},
		{Id: 0x4, EAN: 0x0, Name: "Test4: P1 XN XE T23", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x2, Name: "tag2", Color: 0x1e240}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x5, EAN: 0x0, Name: "Test5: P1 XN XE T1", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}}},
		{Id: 0x6, EAN: 0x0, Name: "Test6: P1 XN XE T13", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x7, EAN: 0x0, Name: "Test7: P1 XN XE T12", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x2, Name: "tag2", Color: 0x1e240}}},
		{Id: 0x8, EAN: 0x0, Name: "Test8: P1 XN XE T123", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x2, Name: "tag2", Color: 0x1e240}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x9, EAN: 0x7b, Name: "Test9: P2 VN XE T", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{}},
		{Id: 0xa, EAN: 0x7c, Name: "Test10: P2 VN XE T3", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0xb, EAN: 0x7d, Name: "Test11: P2 VN XE T2", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x2, Name: "tag2", Color: 0x1e240}}},
		{Id: 0xc, EAN: 0x78, Name: "Test12: P2 VN XE T23", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x2, Name: "tag2", Color: 0x1e240}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0xd, EAN: 0x7f, Name: "Test13: P2 VN XE T1", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}}},
		{Id: 0xe, EAN: 0x80, Name: "Test14: P2 VN XE T13", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0xf, EAN: 0x81, Name: "Test15: P2 VN XE T12", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x2, Name: "tag2", Color: 0x1e240}}},
		{Id: 0x10, EAN: 0x82, Name: "Test16: P2 VN XE T123", TTL: types.TTL{Normal: 0x0, Fridge: 0x0, Freezer: 0x0}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x2, Name: "tag2", Color: 0x1e240}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x11, EAN: 0x0, Name: "Test17: P3 XN VE T", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{}},
		{Id: 0x12, EAN: 0x0, Name: "Test18: P3 XN VE T3", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x13, EAN: 0x0, Name: "Test19: P3 XN VE T2", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x2, Name: "tag2", Color: 0x1e240}}},
		{Id: 0x14, EAN: 0x0, Name: "Test20: P3 XN VE T23", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x2, Name: "tag2", Color: 0x1e240}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x15, EAN: 0x0, Name: "Test21: P3 XN VE T1", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}}},
		{Id: 0x16, EAN: 0x0, Name: "Test22: P3 XN VE T13", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x17, EAN: 0x0, Name: "Test23: P3 XN VE T12", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x2, Name: "tag2", Color: 0x1e240}}},
		{Id: 0x18, EAN: 0x0, Name: "Test24: P3 XN VE T123", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x2, Name: "tag2", Color: 0x1e240}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x19, EAN: 0xea, Name: "Test25: P4 VN VE T", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{}},
		{Id: 0x1a, EAN: 0xeb, Name: "Test26: P4 VN VE T3", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x1b, EAN: 0xec, Name: "Test27: P4 VN VE T2", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x2, Name: "tag2", Color: 0x1e240}}},
		{Id: 0x1c, EAN: 0xed, Name: "Test28: P4 VN VE T23", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x2, Name: "tag2", Color: 0x1e240}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x1d, EAN: 0xee, Name: "Test29: P4 VN VE T1", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}}},
		{Id: 0x1e, EAN: 0xef, Name: "Test30: P4 VN VE T13", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
		{Id: 0x1f, EAN: 0xf0, Name: "Test31: P4 VN VE T12", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x2, Name: "tag2", Color: 0x1e240}}},
		{Id: 0x20, EAN: 0xf1, Name: "Test32: P4 VN VE T123", TTL: types.TTL{Normal: 0x32, Fridge: 0x3c, Freezer: 0x5a}, Tags: []types.Tag{{Id: 0x1, Name: "tag1", Color: 0x0}, {Id: 0x2, Name: "tag2", Color: 0x1e240}, {Id: 0x3, Name: "tag3", Color: 0xfd02ff}}},
	}
)

func TestCreateProduct(t *testing.T) {
	expect := testProducts[0]
	mock := new(mockedDataSource)
	mock.On("InsertProduct", expect).Return(nil)
	lgc := logic.UseDatasource(mock)

	assert.NoError(t, lgc.CreateProduct(expect))
}

func TestCreateProductInsertError(t *testing.T) {
	expect := testProducts[0]
	mock := new(mockedDataSource)
	mock.On("InsertProduct", expect).Return(assert.AnError)
	lgc := logic.UseDatasource(mock)

	err := lgc.CreateProduct(expect)
	require.Error(t, err)
	assert.ErrorIs(t, err, assert.AnError)
}

func TestGetAllProducts(t *testing.T) {
	expect := testProducts
	mock := new(mockedDataSource)
	mock.On("SelectAllProducts").Return(expect, nil)
	lgc := logic.UseDatasource(mock)

	actual, err := lgc.GetAllProducts()
	require.NoError(t, err)
	assert.ElementsMatch(t, actual, expect, tlmIncorrectProducts)
}

func TestGetAllProductsEmpty(t *testing.T) {
	expect := []*types.Product{}
	mock := new(mockedDataSource)
	mock.On("SelectAllProducts").Return(expect, nil)
	lgc := logic.UseDatasource(mock)

	actual, err := lgc.GetAllProducts()
	require.NoError(t, err)
	assert.ElementsMatch(t, actual, expect, tlmIncorrectProducts)
}

func TestGetAllProductsSelectError(t *testing.T) {
	mock := new(mockedDataSource)
	mock.On("SelectAllProducts").Return(nil, assert.AnError)
	lgc := logic.UseDatasource(mock)

	_, err := lgc.GetAllProducts()
	require.Error(t, err)
	assert.ErrorIs(t, err, assert.AnError)
}

func TestGetProductByID(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for edx, expect := range testProducts {
		mock.On("SelectProductById", uint(edx)).Return(expect, nil)
		actual, err := lgc.GetProductByID(uint(edx))
		require.NoError(t, err)
		assert.Equal(t, expect, actual, tlmIncorrectProduct)
	}
}

func TestGetProductByIDNotFound(t *testing.T) {
	mock := new(mockedDataSource)
	mock.On("SelectProductById", uint(0)).Return(&types.Product{}, nil)
	lgc := logic.UseDatasource(mock)

	_, err := lgc.GetProductByID(0)
	require.Error(t, err)
	assert.ErrorIs(t, err, types.ErrIDNotFound)
}

func TestGetProductByIDError(t *testing.T) {
	mock := new(mockedDataSource)
	mock.On("SelectProductById", uint(0)).Return(nil, assert.AnError)
	lgc := logic.UseDatasource(mock)

	_, err := lgc.GetProductByID(0)
	require.Error(t, err)
	assert.ErrorIs(t, err, assert.AnError)
}

func TestGetProductsLike(t *testing.T) {
	ps := &types.ProductSearch{
		Ean:      "241",
		Names:    []string{"Test", "P", "N", "E"},
		TagIds:   []uint{1, 2, 3},
		Oprators: types.ProductSearchOperators{Master: "AND", Names: "AND", TagIds: "AND"},
	}
	mock := new(mockedDataSource)
	mock.On("SelectProductsLike", ps).Return(testProducts[30:], nil)

	lgc := logic.UseDatasource(mock)

	actual, err := lgc.GetProductsLike(ps)
	require.NoError(t, err)
	assert.Equal(t, testProducts[30:], actual, tlmIncorrectProduct)
}

func TestGetProductsLikeEmpty(t *testing.T) {
	ps := &types.ProductSearch{
		Ean:      "241",
		Names:    []string{"Test", "P", "N", "E"},
		TagIds:   []uint{1, 2, 3},
		Oprators: types.ProductSearchOperators{Master: "AND", Names: "AND", TagIds: "AND"},
	}
	mock := new(mockedDataSource)
	mock.On("SelectProductsLike", ps).Return([]*types.Product{}, nil)

	lgc := logic.UseDatasource(mock)

	actual, err := lgc.GetProductsLike(ps)
	require.NoError(t, err)
	assert.Equal(t, []*types.Product{}, actual, tlmIncorrectProduct)
}

func TestGetProductsLikeError(t *testing.T) {
	ps := &types.ProductSearch{
		Ean:      "241",
		Names:    []string{"Test", "P", "N", "E"},
		TagIds:   []uint{1, 2, 3},
		Oprators: types.ProductSearchOperators{Master: "AND", Names: "AND", TagIds: "AND"},
	}
	mock := new(mockedDataSource)
	mock.On("SelectProductsLike", ps).Return(nil, assert.AnError)

	lgc := logic.UseDatasource(mock)

	_, err := lgc.GetProductsLike(ps)
	require.Error(t, err)
	assert.ErrorIs(t, err, assert.AnError)
}

func TestUpdateProduct(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, p := range testProducts {
		mock.On("SelectProductById", p.Id).Return(p, nil)
		mock.On("UpdateProduct", p).Return(p, nil)
		actual, err := lgc.UpdateProduct(p)
		assert.NoError(t, err)
		assert.Equal(t, p, actual)
	}
}

func TestUpdateProductNotFound(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, p := range testProducts {
		mock.On("SelectProductById", p.Id).Return(new(types.Product), nil)
		actual, err := lgc.UpdateProduct(p)
		require.Error(t, err)
		assert.ErrorIs(t, err, types.ErrIDNotFound)
		assert.Equal(t, new(types.Product), actual)
	}
}

func TestUpdateProductSelectError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, p := range testProducts {
		mock.On("SelectProductById", p.Id).Return(nil, assert.AnError)
		actual, err := lgc.UpdateProduct(p)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
		assert.Nil(t, actual)
	}
}

func TestUpdateProductUpdateError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, p := range testProducts {
		mock.On("SelectProductById", p.Id).Return(p, nil)
		mock.On("UpdateProduct", p).Return(nil, assert.AnError)
		actual, err := lgc.UpdateProduct(p)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
		assert.Nil(t, actual)
	}
}

func TestDeleteProduct(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, p := range testProducts {
		mock.On("SelectProductById", uint(p.Id)).Return(p, nil)
		mock.On("DeleteProduct", uint(p.Id)).Return(p, nil)
		actual, err := lgc.DeleteProduct(p.Id)
		require.NoError(t, err)
		assert.Equal(t, p, actual)
	}
}

func TestDeleteProductNotFound(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, p := range testProducts {
		mock.On("SelectProductById", uint(p.Id)).Return(&types.Product{}, nil)
		_, err := lgc.DeleteProduct(p.Id)
		require.Error(t, err)
		assert.ErrorIs(t, err, types.ErrIDNotFound)
	}
}

func TestDeleteProductSelectError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, p := range testProducts {
		mock.On("SelectProductById", uint(p.Id)).Return(nil, assert.AnError)
		_, err := lgc.DeleteProduct(p.Id)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
	}
}

func TestDeleteProductDeleteError(t *testing.T) {
	mock := new(mockedDataSource)
	lgc := logic.UseDatasource(mock)

	for _, p := range testProducts {
		mock.On("SelectProductById", uint(p.Id)).Return(p, nil)
		mock.On("DeleteProduct", uint(p.Id)).Return(nil, assert.AnError)
		_, err := lgc.DeleteProduct(p.Id)
		require.Error(t, err)
		assert.ErrorIs(t, err, assert.AnError)
	}
}
