/**
 * file: tests/logic/utils_test.go
 * author: Theo Technicguy
 * license: Apache 2.0
 *
 * API tests utility functions
 */

package logic_test

import (
	"errors"

	"git.licolas.net/hoffman/server/types"
	"github.com/stretchr/testify/mock"
)

var (
	ErrConversion = errors.New("mock conversion error")
)

// Data source mock object
// This implements the type.datasource
type mockedDataSource struct {
	mock.Mock
}

func (m *mockedDataSource) Disconnect() error {
	// This method exists to implement the data source.
	// This is a connectionless mock that does not need
	// to be connected to or disconnected from.
	return nil
}

func (m *mockedDataSource) CreateHoffmanDatabase() error {
	args := m.Called()
	return args.Error(0)
}

func (m *mockedDataSource) CreateProductTable() error {
	args := m.Called()
	return args.Error(0)
}

func (m *mockedDataSource) DropProductTable() error {
	args := m.Called()
	return args.Error(0)
}

func (m *mockedDataSource) InsertProduct(p *types.Product) error {
	args := m.Called(p)
	return args.Error(0)
}

func (m *mockedDataSource) SelectAllProducts() ([]*types.Product, error) {
	args := m.Called()

	switch v := args.Get(0).(type) {
	case []*types.Product:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) SelectProductById(id uint) (*types.Product, error) {
	args := m.Called(id)

	switch v := args.Get(0).(type) {
	case *types.Product:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) SelectProductByEAN(ean uint) (*types.Product, error) {
	args := m.Called(ean)

	switch v := args.Get(0).(type) {
	case *types.Product:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) SelectProductsLike(ps *types.ProductSearch) ([]*types.Product, error) {
	args := m.Called(ps)

	switch v := args.Get(0).(type) {
	case []*types.Product:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) UpdateProduct(p *types.Product) (*types.Product, error) {
	args := m.Called(p)

	switch v := args.Get(0).(type) {
	case *types.Product:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) DeleteProduct(id uint) (*types.Product, error) {
	args := m.Called(id)

	switch v := args.Get(0).(type) {
	case *types.Product:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) CreateProductTagTable() error {
	args := m.Called()
	return args.Error(0)
}

func (m *mockedDataSource) DropProductTagTable() error {
	args := m.Called()
	return args.Error(0)
}

func (m *mockedDataSource) InsertProductTag(tag *types.Tag) error {
	args := m.Called(tag)
	return args.Error(0)
}

func (m *mockedDataSource) SelectAllProductTags() ([]*types.Tag, error) {
	args := m.Called()

	switch v := args.Get(0).(type) {
	case []*types.Tag:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) SelectProductTagById(id uint) (*types.Tag, error) {
	args := m.Called(id)

	switch v := args.Get(0).(type) {
	case *types.Tag:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) SelectProductTagByName(name string) (*types.Tag, error) {
	args := m.Called(name)

	switch v := args.Get(0).(type) {
	case *types.Tag:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) SelectProductTagsLikeName(name string) ([]*types.Tag, error) {
	args := m.Called(name)

	switch v := args.Get(0).(type) {
	case []*types.Tag:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) UpdateProductTag(tag *types.Tag) (*types.Tag, error) {
	args := m.Called(tag)

	switch v := args.Get(0).(type) {
	case *types.Tag:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}

func (m *mockedDataSource) DeleteProductTag(id uint) (*types.Tag, error) {
	args := m.Called(id)

	switch v := args.Get(0).(type) {
	case *types.Tag:
		return v, args.Error(1)
	case nil:
		return nil, args.Error(1)
	default:
		return nil, ErrConversion
	}
}
